'use strict';

const TOKEN = {
  STYLE_BLOCK_START: "{",
  STYLE_NAME_DELIMITER: " ",
  STYLE_BLOCK_END: "}",
  ESCAPE_CHAR: "\\"
};

const STATE = {
  INIT: Symbol("INIT"),
  TEXT: Symbol("TEXT"),
  IGNORE: Symbol("IGNORE"),
  STYLE_BLOCK_ID: Symbol("STYLE_BLOCK_ID"),
  STYLE_BLOCK_TEXT: Symbol("STYLE_BLOCK_TEXT"),
  STYLE_BLOCK_START: Symbol("STYLE_BLOCK_START"),
  STYLE_BLOCK_END: Symbol("STYLE_BLOCK_END")
};

const TYPE = {
  END_OF_FILE: "END_OF_FILE",
  STYLE_BLOCK_START: "STYLE_BLOCK_START",
  STYLE_BLOCK_ID: "STYLE_BLOCK_ID",
  STYLE_BLOCK_TEXT: "STYLE_BLOCK_TEXT",
  STYLE_BLOCK_END: "STYLE_BLOCK_END",
  TEXT: "TEXT"
};

let lex = "";

// helper functions

function clearLex() {
  lex = "";
}

function addTokenToLex(token) {
  lex += token;
}

// non-deterministic finite automaton

function* scanner(input) {
  let i = 0;
  let state = STATE.INIT;
  clearLex();

  while (true) {
    const token = input[i++];

    if (token === undefined) {
      if (state === STATE.TEXT && lex) {
        yield { type: TYPE.TEXT, lex };
      }

      if (state === STATE.STYLE_BLOCK_TEXT && lex) {
        yield { type: TYPE.STYLE_BLOCK_TEXT, lex };
      }

      return { type: TYPE.END_OF_FILE };
    }

    switch (state) {
      case STATE.INIT:
        {
          if (token === TOKEN.STYLE_BLOCK_START) {
            state = STATE.STYLE_BLOCK_ID;
            yield { type: TYPE.STYLE_BLOCK_START };
            continue;
          } else if (token === TOKEN.ESCAPE_CHAR) {
            state = STATE.IGNORE;
            continue;
          } else {
            state = STATE.TEXT;
            addTokenToLex(token);
            continue;
          }
        }
      case STATE.TEXT:
        {
          if (token === TOKEN.STYLE_BLOCK_START) {
            state = STATE.STYLE_BLOCK_ID;

            yield { type: TYPE.TEXT, lex };
            yield { type: TYPE.STYLE_BLOCK_START };

            clearLex();
            continue;
          }

          if (token === TOKEN.ESCAPE_CHAR) {
            state = STATE.IGNORE;
            continue;
          }

          addTokenToLex(token);
          continue;
        }
      case STATE.IGNORE:
        {
          state = STATE.TEXT;
          lex += token;
          continue;
        }
      case STATE.STYLE_BLOCK_ID:
        {
          if (token === TOKEN.STYLE_NAME_DELIMITER) {
            state = STATE.STYLE_BLOCK_TEXT;

            yield { type: TYPE.STYLE_BLOCK_ID, lex };

            clearLex();
            continue;
          }

          if (token === TOKEN.STYLE_BLOCK_END) {
            state = STATE.TEXT;

            yield { type: TYPE.STYLE_BLOCK_ID, lex };
            yield { type: TYPE.STYLE_BLOCK_END };

            clearLex();
            continue;
          }

          if (token.match(/\s/)) {
            state = STATE.STYLE_BLOCK_TEXT;

            yield { type: TYPE.STYLE_BLOCK_ID, lex };

            clearLex();
            continue;
          }

          lex += token;
          continue;
        }
      case STATE.STYLE_BLOCK_TEXT:
        {
          if (token === TOKEN.STYLE_BLOCK_END) {
            state = STATE.INIT;

            yield { type: TYPE.STYLE_BLOCK_TEXT, lex };
            yield { type: TYPE.STYLE_BLOCK_END };

            clearLex();
            continue;
          }

          addTokenToLex(token);
          continue;
        }
    }
  }
}

const TYPE$1 = {
  TEXT: "TEXT",
  STYLED: "STYLED"
};

let ast = [];
let tokenGenerator = null;

//  helpers functions

function createText(value) {
  ast.push({
    type: TYPE$1.TEXT,
    value
  });
}

function createStyled({ id, value = null }) {
  ast.push({
    type: TYPE$1.STYLED,
    value,
    id
  });
}

function getNextToken() {
  const {
    value: { type, lex }
  } = tokenGenerator.next();

  return { type, lex };
}

// Rules

function Init() {
  const { type, lex } = getNextToken();

  if (type === TYPE.TEXT) {
    createText(lex);

    return Text();
  }

  if (type === TYPE.STYLE_BLOCK_START) {
    return StyledID();
  }

  throw new SyntaxError(`Unexpected token: ${type}`);
}

function StyledID() {
  const { type, lex: id } = getNextToken();

  if (type === TYPE.STYLE_BLOCK_ID) {
    if (!id) {
      throw new SyntaxError(`Unexpected token: ${type}. Token has to have lex.`);
    }

    return StyledText({
      id
    });
  }

  throw new SyntaxError(`Unexpected token: ${type}`);
}

function StyledText(styledObject) {
  const { type, lex: value } = getNextToken();

  if (type === TYPE.STYLE_BLOCK_TEXT) {
    return StyledEnd(Object.assign({
      value
    }, styledObject));
  }

  if (type === TYPE.STYLE_BLOCK_END) {
    createStyled(Object.assign({}, styledObject));

    return Text();
  }

  throw new SyntaxError(`Unexpected token: ${type}`);
}

function StyledEnd(styledObject) {
  const { type, lex } = getNextToken();

  if (type === TYPE.STYLE_BLOCK_END) {
    createStyled(Object.assign({}, styledObject));
    return Text();
  }

  throw new SyntaxError(`Unexpected token: ${type}`);
}

function Text() {
  const { type, lex } = getNextToken();

  if (type === TYPE.STYLE_BLOCK_START) {
    return StyledID();
  }

  if (type === TYPE.TEXT) {
    createText(lex);

    return Text();
  }

  return ast;
}

function parser(input) {
  ast = [];
  tokenGenerator = scanner(input);

  return Init();
}

var _this = undefined;

/**
 * Convert CSS-like style object notation to string
 *
 * Example:
 *
 * convertStyleObjectToString({
 *   fontSize: "20px"
 * }) === "font-size:20;"
 *
 * Numbers are taken as dimensions in pixels
 */
function convertStyleObjectToString(styleObject) {
  return Object.entries(styleObject).map(([rule, value]) => {
    return `${rule.split(/(?=[A-Z])/).join("-").toLowerCase()}:${Number.isInteger(value) ? `${value}px` : value}`;
  }).join(";");
}

/**
 * Call of this function will affect actual console.log call as a side-effect
 *
 * Example:
 *
 * styled("text {styled styled text}", {styled: {fontSize: 20}});
 */
function styled(recipe, stylesDictionary, mapperDictionary) {
  const parsedRecipe = parser(recipe);
  const [texts, styles] = parsedRecipe.reduce((config, currentValue, currentIndex, array) => {
    const { type, value, id } = currentValue;

    let text = value || " ";
    let style = "";

    if (!stylesDictionary && !mapperDictionary) {
      config[0] += text;
      return config;
    }

    if (type === TYPE$1.STYLED) {
      const textMapper = mapperDictionary && mapperDictionary[id];
      if (textMapper) {
        try {
          text = textMapper(value) || " ";
        } catch (e) {
          text = " ";
          throw new Error(`Text mapper for "${id}" has thrown an error: ${e.message}`);
        }
      }

      const styleObject = stylesDictionary && stylesDictionary[id];
      if (styleObject) {
        style = convertStyleObjectToString(styleObject);
      }
    }

    // concat text
    config[0] += `%c${text}`;
    // add new style
    config[1].push(style);
    return config;
  }, ["", []]);

  console.log(texts, ...styles);
}

/**
 * Curried version of "styled" function with reversed order of arguments (data-last)
 *
 * Example:
 *
 * const superConsole = styled.fp(null)({ big: { fontSize: "20px" } });
 * superConsole("{big text}");
 */
styled.fp = (...arg) => {
  return (...x) => {
    const args = [...arg, ...x];

    if (args.length === styled.length) {
      return styled.apply(_this, args.reverse());
    }

    return styled.fp(args);
  };
};

module.exports = styled;
